# Edit this to include the correct information if you want to test against a
# real database

# our $dsn = "DBI:mysql:dod_test:127.0.0.1:3306";
# our $user = "dod_test";
# our $pass = "dod_test";

1 ; # So the require doesn't bork
